"""OAuth 2.0 WSGI server middleware providing MyProxy certificates as access tokens
"""
__author__ = "R B Wilkinson"
__date__ = "28/02/12"
__copyright__ = "(C) 2012 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id$"

class RendererInterface(object):
    """Interface for rendering templated pages.
    """
    def render(self, filename, parameters):
        """Render a page from a template.
        @type filename: basestring
        @param filename: filename of template
        @type parameters: dict
        @param parameters: parameters to substitute into template
        @rtype: basestring
        @return: rendered template
        """
        return None



"""OAuth 2.0 WSGI server middleware providing MyProxy certificates as access tokens
"""
__author__ = "R B Wilkinson"
__date__ = "09/12/11"
__copyright__ = "(C) 2011 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id$"

import json
import logging
import urllib
import uuid

from ndg.httpsclient import utils as httpsclient_utils
from ndg.httpsclient import ssl_context_util

log = logging.getLogger(__name__)


class Oauth2ClientError(Exception):
    '''Base class for OAuth 2.0 client exceptions'''
    
class Oauth2ClientConfigError(Oauth2ClientError):
    '''OAuth 2.0 Client configuration error'''
    
    
class Oauth2ClientConfig(object):
    """OAuth client configuration.
    """

    def __init__(self, client_id, authorization_endpoint, access_token_endpoint,
                 base_url_path, redirect_uri, **kw):
        """
        @type client_id: str
        @param client_id: OAuth client ID

        @type authorization_endpoint: str
        @param authorization_endpoint: URL of OAuth service providing providing
        authorization grants

        @type access_token_endpoint: str
        @param access_token_endpoint: URL of OAuth service providing providing
        access tokens

        @type base_url_path: str
        @param base_url_path: base path included in OAuth client URLs

        @type redirect_uri: str
        @param redirect_uri: URL to which the OAuth authorization server should
        redirect after an authorization request

        @type kw: dict
        @param kw: additional parameters for configuring extended client classes
        """
        self.client_id = client_id
        self.authorization_endpoint = authorization_endpoint
        self.access_token_endpoint = access_token_endpoint
        self.base_url_path = base_url_path
        self.redirect_uri = redirect_uri
        self.kw = kw

    def make_redirect_uri(self, application_url):
        """Constructs the redirect URI from components.

        @type application_url: str
        @param application_url: application base URL

        @rtype: str
        @return: full redirect URI
        """
        log.debug("make_redirect_uri: application_url=%s base_url_path=%s "
                  "redirect_uri=%s",
                  application_url, self.base_url_path, self.redirect_uri)
        return self._construct_url([application_url, self.base_url_path,
                                    self.redirect_uri])

    def is_redirect_uri(self, application_url, url):
        """Determines whether a URL is the redirect URI for this client.

        @type application_url: str
        @param application_url: application base URL

        @type url: str
        @param url: URL to check

        @rtype: bool
        @return: True if URL is the redirect URI otherwise False
        """
        redirect_uri = self._construct_url([application_url, self.base_url_path,
                                            self.redirect_uri])
        path_url = url.partition('?')[0]
        return redirect_uri == path_url

    @staticmethod
    def _construct_url(parts):
        """Joins components of a URL with separators where needed.

        @type parts: dict of str
        @param parts: URL components

        @rtype: str
        @return: combined URL (part)
        """
        results = []
        last_idx = len(parts) - 1
        for n, part in enumerate(parts):
            if n > 0:
                part = part.lstrip('/')
            if n < last_idx:
                part = part.rstrip('/')
            if part:
                results.append(part)
        return '/'.join(results)


class Oauth2Client(object):
    """OAuth 2.0 client
    """
    ACCESS_TOK_ENCODING = 'utf-8'
    RESPONSE_TYPE = 'code'
    SESSION_ID_KEY = 'oauth_client_instance_id'
    HTTP_AUTHORIZATION_HEADER_FIELD = 'Authorization'
    BEARER_TOK_ID = 'Bearer'
    MAC_TOK_ID = 'MAC'
    TOKEN_TYPES = (BEARER_TOK_ID, MAC_TOK_ID)

    def __init__(self, client_config=None, access_token=None):
        """
        @type client_config: ndgoauthclient.lib.oauth2client.Oauth2ClientConfig
        @param client_config: OAuth client configuration
        """
        self.client_config = client_config
        self.access_token = access_token
        
        if client_config is not None:
            for k, v in client_config.kw.iteritems():
                setattr(self, k, v)
        
    def call_with_access_token(self, scope, application_url, callback):
        """Calls a specified callable providing an access token.
        
        @type scope: str
        @param scope: required OAuth token scope

        @type application_url: str
        @param application_url: application base URL

        @type callback: callable called with arguments
            (access_token, error, error_description)
        @param callback: callable to call when the token is available

        @rtype: tuple (
            result (None if not obtained yet)
            redirect URL (None if access token already obtained)
        )
        @return: return value from callback if access token available or a
            URL to which to redirect to obtain an access token
        """
        if self.access_token is not None:
            log.debug("call_with_access_token: token found")
            result = callback(self.access_token, None, None)
            return (result, None)
        log.debug("call_with_access_token: token not found")

        # Client does not have an access token, so create redirect URI with
        # which to initiate the process of getting one.
        redirect_uri = self.client_config.make_redirect_uri(application_url)
        parameters = {
            'client_id': self.client_config.client_id,
            'redirect_uri': redirect_uri,
            'response_type': self.RESPONSE_TYPE,
            'scope': scope
            }
        self.state = uuid.uuid4().hex
        log.debug("call_with_access_token authorization_endpoint: %s", 
                  self.client_config.authorization_endpoint)
        log.debug("call_with_access_token parameters: %s", parameters)
        url = self._make_combined_url(self.client_config.authorization_endpoint, 
                                      parameters, self.state)
        return (None, url)

    def call_with_access_token_redirected_back(self, request, callback,
                                               ssl_config):
        """Called after redirection following authorization process.

        @type request: webob.Request
        @param request: request object

        @type callback: callable called with arguments
            (access_token, error, error_description)
        @param callback: callable to call when the token is available

        @type ssl_config: ndg.httpsclient.ssl_context_util.SSlContextConfig
        @param ssl_config: SSL configuration

        @rtype: any
        @return: result from callback
        """
        params = request.GET
        code = params.get('code', None)
        state = params.get('state', None)
        error = params.get('error', None)
        error_description = params.get('error_description', None)
        if error:
            log.info("Error from OAuth authorization server: %s - %s",
                     error, error_description)
            return callback(None, error, error_description)
        else:
            if state != self.state:
                error = 'Inconsistent state'
                error_description = (
                    'State value incorrect implying request is not the result '
                    'of legitimate OAuth redirection.')
                
                log.info("Erroneous request to redirect URL: %s - %s",
                         error, error_description)
                return callback(None, error, error_description)
            else:
                log.debug("Valid redirect from OAuth authorization server")
                return self.request_access_token(code, request.application_url,
                                                 request, callback, ssl_config)

    def request_access_token(self, code, application_url, request, callback,
                             ssl_config):
        """
        @type code: str
        @param code: authorization code

        @type application_url: str
        @param application_url: application base URL

        @type request: webob.Request
        @param request: request object

        @type callback: callable called with arguments
            (access_token, error, error_description)
        @param callback: callable to call when the token is available

        @type ssl_config: ndg.httpsclient.ssl_context_util.SSlContextConfig
        @param ssl_config: SSL configuration

        @rtype: any
        @return: result from callback
        """
        # Make POST request to obtain an access token.
        redirect_uri = self.client_config.make_redirect_uri(application_url)
        parameters = {
            'grant_type': 'authorization_code',
            'code': code,
            'redirect_uri': redirect_uri
        }
        
        # Put client_secret in request when defined. Most implementations
        # require this as part of the parameters, even though using an
        # authorization header with a bearer-token would seem to be preferred
        # by draft-ietf-oauth-v2-bearer-23
        if self.client_config.kw.get('client_secret') is not None:
            parameters['client_id'] = self.client_config.client_id
            parameters['client_secret'] = self.client_config.kw['client_secret']

        self.additional_access_token_request_parameters(parameters, request)
        
        log.debug("Requesting access token - parameters: %s", parameters)
        data = urllib.urlencode(parameters)
        
        ssl_ctx = ssl_context_util.make_ssl_context_from_config(ssl_config)
        
        # Header required by, for example Github, to get a json response
        config = httpsclient_utils.Configuration(
                                        ssl_ctx,
                                        headers={'Accept': 'application/json'})
            
        response_json = httpsclient_utils.fetch_stream_from_url(
                                    self.client_config.access_token_endpoint,
                                    config,
                                    data)
        
        response = json.load(response_json)
        access_token = response.get('access_token', None)
        if 'error' in response:
            error = response['error']
            error_description = response.get('error_description', None)
            
            log.error('Access token request error: %s %s', error, 
                      error_description)
            
            return callback(None, error, error_description)
        
        elif access_token is None:
            error = 'invalid_request'
            error_description = ('Error retrieving access token - '
                                 'no access token returned.')
            
            return callback(None, error, error_description)
        else:
            self.access_token = access_token.encode(self.ACCESS_TOK_ENCODING)
            
            log.debug("Access token received: %s", self.access_token)
            
            return callback(self.access_token, None, None)

    def additional_access_token_request_parameters(self, parameters, request):
        """
        Override to add parameters into the access token request.

        @type parameters: dict of str
        @param parameters: parameters sent in access token request

        @type request: webob.Request
        @param request: request object
        """

    def request_resource(self, 
                         resource_url, 
                         ssl_config=None, 
                         config=None, 
                         data=None, 
                         handlers=None):
        '''Request a resource URL setting *bearer* token in Authorization
        header

        @type resource_url: str
        @param resource_url: URL of resource to be requested

        @type ssl_config: ndg.httpsclient.ssl_context_util.SSlContextConfig
        @param ssl_config: SSL configuration, Nb. if config keyword is set then
        this parameter will be *ignored*
        
        @type config: ndg.httpsclient.utils.Configuration
        @param config: HTTP configuration settings.  Setting this keyword will
        override any setting made for ssl_config      
        @param data: HTTP POST data
        @type data: str
        @param handlers: list of custom urllib2 handlers to add to the request
        @type handlers: iterable
        @return: response from resource server
        @rtype: urllib.addinfourl
        '''
        if self.access_token is None:
            raise Oauth2ClientConfigError('No access token set for request to '
                                          'resource %r' % resource_url)
        
        authorization_header = {
            self.__class__.HTTP_AUTHORIZATION_HEADER_FIELD: '%s %s' % (
                                                self.__class__.BEARER_TOK_ID, 
                                                self.access_token),
        }
        
        if config and ssl_config:
            raise TypeError('Set either "config" or "ssl_config" keywords but '
                            'not both')
            
        if config is None:
            ssl_ctx = ssl_context_util.make_ssl_context_from_config(ssl_config)
            config = httpsclient_utils.Configuration(
                                                ssl_ctx,
                                                headers=authorization_header)
        
        response = httpsclient_utils.fetch_stream_from_url(resource_url, 
                                                           config,
                                                           data=data)
        return response


    @staticmethod
    def _make_combined_url(base_url, parameters, state):
        """Combines a base URL and parameters to make a full URL.

        @type base_url: str
        @param base_url: base URL to which to add query parameters

        @type parameters: dict of str
        @param parameters: name/values pairs to use as query parameters

        @type state: str
        @param state: state parameter which must be included without encoding

        @rtype: str
        @return: combined URL
        """
        url = base_url.rstrip('?')
        url_parts = [url]
        sep_with_ampersand = ('?' in url)
        if parameters:
            query_string = urllib.urlencode(parameters)
            url_parts.extend([('&' if (sep_with_ampersand) else '?'), 
                              query_string])
            sep_with_ampersand = True

        if state:
            url_parts.extend([('&' if (sep_with_ampersand) else '?'), 
                              'state=',
                              state])

        return ''.join(url_parts)

    @classmethod
    def get_client_instance(cls, session, client_config, create=False):
        """Retrieves from the session, and optionally creating if necessary, a
        client instance.

        @type session: SessionObject
        @param session: session data

        @type client_config: 
        @param client_config: client configuration

        @type create: bool
        @param create: True if a client is to be created if it does not exist
            already

        @rtype: instance of Oauth2Client
        @return: OAuth 2 client object
        """
        client = None
        if cls.SESSION_ID_KEY in session:
            client = session[cls.SESSION_ID_KEY]
            log.debug("Found OAuth client in session.")
        if client is None and create:
            client = cls(client_config)
            session[cls.SESSION_ID_KEY] = client
            session.save()
            log.debug("No OAuth client in session - created new one.")
        return client

    @classmethod
    def delete_client_instance(cls, session):
        """Deletes the client instance from the session.

        @type session: SessionObject
        @param session: session data
        """
        if cls.SESSION_ID_KEY in session:
            client_instance_id = session[cls.SESSION_ID_KEY]
            
            if client_instance_id in cls.client_instances:
                del cls.client_instances[client_instance_id]



"""OAuth 2.0 WSGI server middleware providing MyProxy certificates as access tokens
"""
__author__ = "R B Wilkinson"
__date__ = "09/12/11"
__copyright__ = "(C) 2011 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id$"

import base64

from ndg.oauth.client.lib.oauth2client import Oauth2Client
import ndg.oauth.client.lib.openssl_cert as openssl_cert

class Oauth2MyProxyClient(Oauth2Client):
    """Extentsion of OAuth client to handle additional certificate request
    parameter for MyProxy request.
    """
    def additional_access_token_request_parameters(self, parameters, request):
        """
        Creates a certificate request and sets it as an additional parameter.
        Also, saves the private key in environ.

        @type parameters: dict of str
        @param parameters: parameters sent in access token request

        @type request: webob.Request
        @param request: request object
        """
        key_pair = openssl_cert.create_keypair()
        cert_req = openssl_cert.create_certreq('ignored-username', key_pair)
        parameters[self.certificate_request_parameter] = base64.b64encode(cert_req)

        # Store the private key.
        private_key = openssl_cert.getKeyPairPrivateKey(key_pair)
        self.private_key = private_key



"""OAuth 2.0 WSGI server middleware providing MyProxy certificates as access tokens
"""
__author__ = "R B Wilkinson"
__date__ = "09/12/11"
__copyright__ = "(C) 2011 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id$"

import httplib
import logging

from webob import Request

from ndg.httpsclient.ssl_context_util import SSlContextConfig
from ndg.oauth.client.lib.oauth2client import Oauth2Client
from ndg.oauth.client.lib.oauth2client import Oauth2ClientConfig
from ndg.oauth.client.lib.oauth2_myproxy_client import Oauth2MyProxyClient
from ndg.oauth.client.lib.render.configuration import RenderingConfiguration
from ndg.oauth.client.lib.render.factory import callModuleObject
from ndg.oauth.client.lib.render.renderer_interface import RendererInterface

log = logging.getLogger(__name__)


class Oauth2ClientMiddlewareSessionError(Exception):
    """Missing beaker session key"""
    
    
class Oauth2ClientMiddlewareConfigOptError(Exception):
    """Error with Config option"""
    
    
class Oauth2ClientMiddleware(object):
    """
    OAuth 2.0 client middleware that sets an access token in the WSGI environ.
    """
    PARAM_PREFIX = 'oauth2.'
    LAYOUT_PREFIX = 'layout.'
    ACCESS_TOKEN_TYPE_OPTION = 'access_token_type'
    AUTHENTICATION_TRIGGER_OPTION = 'authentication_trigger'
    AUTHENTICATION_TRIGGER_ALWAYS = 'always'
    AUTHENTICATION_TRIGGER_UNAUTHORIZED = 'unauthorized'
    AUTHENTICATION_TRIGGER_URL = 'url'
    AUTHENTICATION_TRIGGER_OPTIONS = [AUTHENTICATION_TRIGGER_ALWAYS,
                                      AUTHENTICATION_TRIGGER_UNAUTHORIZED,
                                      AUTHENTICATION_TRIGGER_URL]
    AUTHENTICATION_URL_OPTION = 'authentication_url'
    AUTHENTICATION_COMPLETE_OPTION = 'login_complete'
    BASE_URL_PATH_OPTION = 'base_url_path'
    CERTIFICATE_REQUEST_PARAMETER_OPTION = 'certificate_request_parameter'
    REDIRECT_URI = 'oauth_redirect'
    RENDERER_CLASS_OPTION = 'renderer_class'
    SCOPE_OPTION = 'scope'
    SESSION_KEY_OPTION = 'session_key'
    SESSION_CALL_CONTEXT_KEY = 'oauth2_call_context'
    TOKEN_KEY_OPTION = 'oauth2_token_key'
    CLIENT_CERT_OPTION = 'client_cert'
    CLIENT_KEY_OPTION = 'client_key'
    CA_CERT_FILE_OPTION = 'ca_cert_file'
    CA_DIR_OPTION = 'ca_dir'
    CLIENT_ID_OPTION = 'client_id'
    CLIENT_SECRET_OPTION = 'client_secret'
    AUTHORIZATION_ENDPOINT_OPTION = 'authorization_endpoint'
    ACCESS_TOKEN_ENDPOINT_OPTION = 'access_token_endpoint'
    
    PROPERTY_DEFAULTS = {
        ACCESS_TOKEN_TYPE_OPTION: 'bearer',
        AUTHENTICATION_COMPLETE_OPTION: '',
        AUTHENTICATION_TRIGGER_OPTION: AUTHENTICATION_TRIGGER_ALWAYS,
        AUTHENTICATION_URL_OPTION: 'oauth_authenticate',
        BASE_URL_PATH_OPTION: '',
        CERTIFICATE_REQUEST_PARAMETER_OPTION: 'certificate_request',
        RENDERER_CLASS_OPTION: \
            'ndg.oauth.client.lib.render.genshi_renderer.GenshiRenderer',
        SCOPE_OPTION: '',
        SESSION_KEY_OPTION: 'beaker.session.oauth2client',
        TOKEN_KEY_OPTION: 'oauth2client.token',
        CLIENT_SECRET_OPTION: None,
    }
    LAYOUT_PARAMETERS = ['heading',
                         'title',
                         'message',
                         'leftLogo',
                         'leftAlt',
                         'leftImage',
                         'leftLink',
                         'rightAlt',
                         'rightImage',
                         'rightLink',
                         'footerText',
                         'helpIcon',
                         'client_id',
                         'client_secret',
                         'client_name',
                         'scope']

    client_instances = {}

    def __init__(self, app, app_conf, prefix=PARAM_PREFIX, **local_conf):
        """
        @param app: wrapped application/middleware
        @type app: WSGI application
        @param app_conf: application configuration settings - ignored - this
        method includes this arg to fit Paste middleware / app function 
        signature
        @type app_conf: dict
        @param prefix: optional prefix for parameter names included in the 
        local_conf dict - enables these parameters to be filtered from others
        which don't apply to this middleware
        @param local_conf: attribute settings to apply
        @type local_conf: dict
        """
        self._app = app
        self._set_configuration(prefix, local_conf)
        if self.access_token_type == 'slcs':
            log.debug("Setting client as Oauth2MyProxyClient (SLCS)")
            self._oauth_client_class = Oauth2MyProxyClient
            self._token_retriever_class = MyProxyTokenRetriever
        else:
            log.debug("Setting client as Oauth2Client (Bearer token)")
            self._oauth_client_class = Oauth2Client
            self._token_retriever_class = TokenRetriever

        self._renderingConfiguration = RenderingConfiguration(
                                                    self.LAYOUT_PARAMETERS,
                                                    prefix + self.LAYOUT_PREFIX,
                                                    local_conf)
        self.renderer = callModuleObject(self.renderer_class,
                                         objectName=None, moduleFilePath=None, 
                                         objectType=RendererInterface,
                                         objectArgs=None, objectProperties=None)

    def __call__(self, environ, start_response):
        """
        @param environ: WSGI environment
        @type environ: dict
        @param start_response: WSGI start response function
        @type start_response: 
        @return: WSGI response
        @rtype: iterable
        """
        log.debug("Oauth2ClientMiddleware.__call__ ...")

        req = Request(environ)
        log.debug("Request url: %s", req.url)
        log.debug("Request host_url: %s", req.host_url)
        log.debug("Request application_url: %s", req.application_url)
        is_redirect_back = False
        
        original_environ = {'PATH_INFO': environ['PATH_INFO'],
                            'QUERY_STRING': environ['QUERY_STRING'],
                            'SCRIPT_NAME': environ['SCRIPT_NAME'],
                            'url': req.url}

        # Get session.
        session = environ.get(self.session_env_key)
        if session is None:
            raise Oauth2ClientMiddlewareSessionError(
                'Oauth2ClientMiddleware.__call__: No beaker session key '
                '"%s" found in environ' % self.session_env_key)

        # Determine trigger for starting authentication process.
        authenticate_before_delegating = False
        authenticate_on_unauthorized = False
        is_authentication_url = (environ['PATH_INFO'].strip('/') ==
                                                        self.authentication_url)
        if (self.authentication_trigger ==
            self.__class__.AUTHENTICATION_TRIGGER_ALWAYS):
            authenticate_before_delegating = True
            
        elif (self.authentication_trigger ==
              self.__class__.AUTHENTICATION_TRIGGER_URL):
            
            if is_authentication_url:
                authenticate_before_delegating = True
                
        elif (self.authentication_trigger ==
              self.__class__.AUTHENTICATION_TRIGGER_UNAUTHORIZED):
            authenticate_on_unauthorized = True

        # Check whether redirecting back after requesting authorization.
        redirect_url = None
        if self.client_config.is_redirect_uri(req.application_url, req.url):
            log.debug("Redirected back after requesting authorization.")
            is_redirect_back = True
            token = self._get_token_after_redirect(session, req)
            original_environ = session[self.__class__.SESSION_CALL_CONTEXT_KEY]
        else:
            # Start the OAuth2 transaction to get a token.
            log.debug("Starting OAuth2 protocol")
            token, redirect_url = self._get_token(session,
                                                  req.application_url)
            if authenticate_before_delegating and redirect_url:
                session[self.__class__.SESSION_CALL_CONTEXT_KEY
                                                            ] = original_environ
                session.save()
                log.debug("Redirecting to %s", redirect_url)
                start_response(self._get_http_status_string(httplib.FOUND),
                               [('Location', redirect_url)])
                return []

        local_start_response = start_response
        if token:
            log.debug("Setting token in environ[%s]=%s", self.token_env_key,
                      token)
            environ[self.token_env_key] = token
            
        elif authenticate_on_unauthorized and redirect_url:
            def local_start_response(status, response_headers, exc_info=None):
                status_code = status.split(' ')[0]
                log.debug("Response HTTP status %s", status_code)
                if status_code == str(httplib.UNAUTHORIZED):
                    session[self.__class__.SESSION_CALL_CONTEXT_KEY
                                                            ] = original_environ
                    session.save()
                    log.debug("Redirecting to %s", redirect_url)
                    start_response(self._get_http_status_string(httplib.FOUND),
                                   [('Location', redirect_url)])
                    return []
                else:
                    return start_response(status, response_headers, exc_info)

        if is_authentication_url:
            c = {'baseURL': req.application_url}
            response = self.renderer.render(self.authentication_complete,
                            self._renderingConfiguration.merged_parameters(c))
            
            start_response(self._get_http_status_string(httplib.OK),
                           [('Content-type', 'text/html'),
                            ('Content-length', str(len(response)))
                            ])
            
            return [response]

        # Ensure that the URL is that prior to authentication redirection.
        if is_redirect_back:
            original_url = original_environ['url']
            log.debug("Redirecting to %s", original_url)
            start_response(self._get_http_status_string(httplib.FOUND),
                           [('Location', original_url)])
            return []

        app_iter = self._app(environ, local_start_response)
        return app_iter

    def _set_configuration(self, prefix, local_conf):
        """Sets the configuration values.

        @param prefix: optional prefix for parameter names included in the
        local_conf dict - enables these parameters to be filtered from others
        which don't apply to this middleware
        @type prefix: str
        @param local_conf: attribute settings to apply
        @type local_conf: dict
        """
        cls = self.__class__
        self.access_token_type = cls._get_config_option(prefix, local_conf,
                                                cls.ACCESS_TOKEN_TYPE_OPTION)
        self.authentication_complete = cls._get_config_option(prefix,
                                local_conf, cls.AUTHENTICATION_COMPLETE_OPTION)
        
        self.authentication_trigger = cls._get_config_option(
                prefix, local_conf, cls.AUTHENTICATION_TRIGGER_OPTION).lower()
                
        if self.authentication_trigger not in cls.AUTHENTICATION_TRIGGER_OPTIONS:
            raise Oauth2ClientMiddlewareConfigOptError(
                        "Illegal value for %s option; expected one of %s" %
                        self.authentication_trigger_str,
                        cls.AUTHENTICATION_TRIGGER_OPTIONS)
            
        self.authentication_url = cls._get_config_option(
                prefix, local_conf, cls.AUTHENTICATION_URL_OPTION).strip('/')
                
        self.renderer_class = cls._get_config_option(prefix, local_conf, 
                                                     cls.RENDERER_CLASS_OPTION)
        self.scope = cls._get_config_option(prefix, local_conf, 
                                            cls.SCOPE_OPTION)
        self.session_env_key = cls._get_config_option(prefix, local_conf, 
                                                      cls.SESSION_KEY_OPTION)
        self.token_env_key = self._get_config_option(prefix, local_conf, 
                                                     cls.TOKEN_KEY_OPTION)

        # SSL configuration
        client_cert = cls._get_config_option(prefix, local_conf, 
                                             cls.CLIENT_CERT_OPTION)
        client_key = cls._get_config_option(prefix, local_conf, 
                                            cls.CLIENT_KEY_OPTION)
        ca_cert_file = cls._get_config_option(prefix, local_conf, 
                                              cls.CA_CERT_FILE_OPTION)
        ca_dir = cls._get_config_option(prefix, local_conf, cls.CA_DIR_OPTION)
        self.ssl_config = SSlContextConfig(client_key, client_cert,
                                           ca_cert_file, ca_dir, True)

        # OAuth client configuration
        certificate_request_parameter = cls._get_config_option(prefix, 
                                    local_conf, 
                                    cls.CERTIFICATE_REQUEST_PARAMETER_OPTION)
        client_id = cls._get_config_option(prefix, local_conf, 
                                           cls.CLIENT_ID_OPTION)
        client_secret = cls._get_config_option(prefix, local_conf, 
                                               cls.CLIENT_SECRET_OPTION)
        authorization_endpoint = cls._get_config_option(prefix, local_conf, 
                                            cls.AUTHORIZATION_ENDPOINT_OPTION)
        access_token_endpoint = cls._get_config_option(prefix, local_conf, 
                                            cls.ACCESS_TOKEN_ENDPOINT_OPTION)
        base_url_path = cls._get_config_option(prefix, local_conf, 
                                            cls.BASE_URL_PATH_OPTION)
        
        redirect_uri = cls.REDIRECT_URI
        
        self.client_config = Oauth2ClientConfig(
                    client_id, 
                    authorization_endpoint, 
                    access_token_endpoint,
                    base_url_path, 
                    redirect_uri,
                    client_secret=client_secret,
                    certificate_request_parameter=certificate_request_parameter)

    @classmethod
    def _get_config_option(cls, prefix, local_conf, key):
        value = local_conf.get(prefix + key, 
                               cls.PROPERTY_DEFAULTS.get(key, None))
        log.debug("Oauth2ClientMiddleware configuration %s=%s", key, value)
        return value

    @staticmethod
    def _get_http_status_string(status):
        return "%d %s" % (status, httplib.responses[status])

    @classmethod
    def filter_app_factory(cls, app, app_conf, **local_conf):
        return cls(app, app_conf, **local_conf)

    def _get_token(self, session, application_url):
        """Gets a token using the OAuth2 client.
        @type session: Beaker SessionObject
        @param session: session
        @type application_url: str
        @param application_url: application base part of request URL
        @rtype: tuple (
            result type of callback or None
            str or None
        )
        @return: (
            result of callback or None if a redirect is needed
            redirect URI if redirect needed or None
        """
        client = self._oauth_client_class.get_client_instance(session,
                                                              self.client_config,
                                                              create=True)

        callback = self._token_retriever_class(client)

        result, redirect_url = client.call_with_access_token(self.scope, 
                                                             application_url, 
                                                             callback)
        session.save()

        return (result, redirect_url)

    def _get_token_after_redirect(self, session, req):
        """Gets a token using the OAuth2 client - to be called after a redirect
        has occurred from the OAuth authorization server.
        @type session: Beaker SessionObject
        @param session: session
        @rtype: result type of callback
        @return: result of callback
        """
        client = self._oauth_client_class.get_client_instance(
                                                            session,
                                                            self.client_config)
        if client:
            # Return callback result.
            callback = self._token_retriever_class(client)
            result = client.call_with_access_token_redirected_back(req,
                                                                callback,
                                                                self.ssl_config)
            # Save client state, which includes the token.
            session.save()
            # Save only marks the session for persistence at the end of the HTTP
            # transaction. Persist now so that it is available if a new request
            # is made from nested middleware.
            session.persist()

            return result
        else:
            raise Oauth2ClientMiddlewareSessionError("No OAuth client created "
                                                     "for session.")


class TokenRetriever(object):
    def __init__(self, client):
        self.client = client

    def __call__(self, access_token, error, error_description):
        """
        Returns authorization token.
        @type access_token: type of access token
        @param access_token: access token
        @type error: str
        @param error: OAuth error string
        @type error_description: str
        @param error_description: error description
        @rtype: type of access token
        @return: access token
        """
        if error:
            return None
        return self.client.access_token


class MyProxyTokenRetriever(object):
    def __init__(self, client):
        self.client = client

    def __call__(self, access_token, error, error_description):
        """
        Returns the private key and certificate.
        This depends on Oauth2MyProxyClient which sets the private key in the
        client.
        @type access_token: type of access token
        @param access_token: access token
        @type error: str
        @param error: OAuth error string
        @type error_description: str
        @param error_description: error description
        @rtype: tuple (str, str)
        @return: tuple (
            private key
            access token
        )
        """
        if error:
            return None
#            return ("", ("Token not available because of error: %s - %s" % (error, error_description)))
        return self.client.private_key, self.client.access_token



"""OAuth 2.0 WSGI server test application for use with client middleware
"""
__author__ = "R B Wilkinson"
__date__ = "09/12/11"
__copyright__ = "(C) 2011 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id$"
import urllib
import cgi
import logging
from abc import ABCMeta, abstractmethod

from ndg.httpsclient import ssl_context_util 
import ndg.httpsclient.utils as httpsclient_utils

from ndg.oauth.client.lib import certificate_request
from ndg.oauth.client.lib.oauth2client import Oauth2Client

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class OAuthClientTestAppBase(object):
    """
    Simple WSGI application that displays a token set in the WSGI environ by
    Oauth2ClientMiddleware.
    """
    __metaclass__ = ABCMeta
    
    METHOD = {
        "/": 'default',
        "/token": 'tok',
        '/resource': 'get_resource'
    }
    
    DEFAULT_TOKEN_ENV_KEYNAME = 'oauth2client.token'
    
    def __init__(self, app, app_conf, **local_conf):
        self.beaker_session_keyname = app_conf['beaker_session_keyname']
        self.app = app

        self.resource_url = local_conf['resource_url']
        self.token_env_key = local_conf.get('oauth2_token_key',
                                            self.DEFAULT_TOKEN_ENV_KEYNAME)

        # SSL configuration
        client_cert = local_conf.get('client_cert')
        client_key = local_conf.get('client_key')
        self.ca_cert_file = local_conf.get('ca_cert_file')
        self.ca_dir = local_conf.get('ca_dir')
        self.ssl_config = ssl_context_util.SSlContextConfig(client_key,
            client_cert, self.ca_cert_file, self.ca_dir, True)
        
        self.oauth2client = Oauth2Client(None)

    def __call__(self, environ, start_response):
        method_name = self.METHOD.get(environ['PATH_INFO'], '').rstrip()
        if method_name:
            action = getattr(self, method_name)
            return action(environ, start_response)
        elif self.app is not None:
            return self.app(environ, start_response)
        else:
            start_response('404 Not Found', 
                           [('Content-type', 'text/plain; charset=UTF-8')])
            return "ndg_oauth WSGI Test Application: invalid URI"

    def default(self, environ, start_response):
        log.debug('%s.default ...', self.__class__.__name__)
        response = """<h2>ndg_oauth WSGI Test Application</h2>
        <ul>
        <li><a href="/token">Display access token</a></li>
        <li><a href="/resource">Get example resource using access token</a></li>
        <ul>
        <p>
        <em>
        Nb. clear out cookies to reset application and re-initiate delegation
        process
        </em>
        </p>
        """
        start_response('200 OK', 
                       [('Content-type', 'text/html; charset=UTF-8'),
                        ('Content-length', str(len(response)))])
        return [response]

    def tok(self, environ, start_response):
        log.debug('%s.tok ...', self.__class__.__name__)

        tok = environ.get(self.__class__.DEFAULT_TOKEN_ENV_KEYNAME)
        response = ["<h2>ndg_oauth WSGI Test Application: Get Token"
                    "</h2>"]
        if tok:
            if isinstance(tok, basestring):
                response.append("<pre>%s</pre>" % tok)
            else:
                for i in tok:
                    response.append("<pre>%s</pre>" % i)
        else:
            response.append("<p>token not found</p>")

        start_response('200 OK', 
                       [('Content-type', 'text/html; charset=UTF-8'),
                        ('Content-length', 
                         str(sum([len(r) for r in response])))])
        return response

    @classmethod
    def app_factory(cls, app_conf, **local_conf):
        return cls(None, app_conf, **local_conf)

    @classmethod
    def filter_app_factory(cls, app, app_conf, **local_conf):
        return cls(app, app_conf, **local_conf)
        
    @abstractmethod
    def get_resource(self, environ, start_response):
        '''Retrieve a resource secured by the OAuth resource server'''
        

class BearerTokOAuthClientApp(OAuthClientTestAppBase):   
    '''Example client application for bearer token OAuth'''
    
    def get_resource(self, environ, start_response):
        response = [
            "<h2>ndg_oauth WSGI Test Application: get secured resource</h2>",
            "<p>Retrieved resource output from %r:<p>" % self.resource_url
        ]

        self.oauth2client.access_token = environ[self.token_env_key]
    
        log.debug("Resource request - access_token: %s", 
                  self.oauth2client.access_token)
     
        resource_server_response = self.oauth2client.request_resource(
                                                    self.resource_url, 
                                                    ssl_config=self.ssl_config)

        output = resource_server_response.read()
        esc_output = cgi.escape(output, True)
        enc_output = esc_output.encode('ascii', 'xmlcharrefreplace')
        response.append('<p>%r</p>' % enc_output)

        start_response('200 OK',
                       [('Content-type', 'text/html; charset=UTF-8'),
                        ('Content-length', str(sum([len(r) for r in response])))
                        ])
        return response
    
            
class SlcsExampleOAuthClientApp(OAuthClientTestAppBase):
    '''Extend basic OAuth client demonstration application to illustrate
    retrieving a a certificate from a Short-Lived Credential Service.
    '''
    DEFAULT_CERTIFICATE_REQUEST_PARAMETER = 'certificate_request'
    
    def __init__(self, app, app_conf, **local_conf):
        super(SlcsExampleOAuthClientApp, self).__init__(app, app_conf, 
                                                        **local_conf)
        
        # OAuth client configuration
        self.certificate_request_parameter = local_conf.get(
                                    'certificate_request_parameter',
                                    self.DEFAULT_CERTIFICATE_REQUEST_PARAMETER)
    
    def get_resource(self, environ, start_response):
        response = [
            "<h2>ndg_oauth WSGI Test Application: get secured resource</h2>"
        ]

        token = environ.get(self.token_env_key)
        if not token:
            response.append('<p>No OAuth token found.</p>')
        else:
            log.debug("Token found; %s" % token)
            try:
                (private_key,
                 certificate) = certificate_request.request_certificate(
                    token, self.resource_url, self.ssl_config,
                    self.certificate_request_parameter)
            except Exception, exc:
                response.append('<p>Exception obtaining certificate from OAuth '
                                'server: %s</p>' % exc)
            else:
                response.append('<p>%s</p><p>%s</p>' % (private_key,
                                                        certificate))

        start_response('200 OK',
                       [('Content-type', 'text/html; charset=UTF-8'),
                        ('Content-length', str(sum([len(r) for r in response])))
                        ])
        return response




"""Retriever of data from a URL using credentials obtained using an OAuth
    access token.
"""
__author__ = "R B Wilkinson"
__date__ = "29/03/12"
__copyright__ = "(C) 2012 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id$"

import logging

from OpenSSL import crypto

import ndg.httpsclient.utils as httpsclientutils
import ndg.httpsclient.ssl_context_util as ssl_context_util
import ndg.oauth.client.lib.certificate_request as certificate_request

log = logging.getLogger(__name__)

class HttpGetWithCredential(object):
    """Retrieves data from a URL using credentials obtained using an OAuth
    access token. 
    """
    DEFAULT_TOKEN_ENV_KEYNAME = 'oauth2client.token'

    def __init__(self, 
                 resource_server_url, 
                 client_cert, 
                 client_key,
                 ca_cert_file, 
                 ca_dir,
                 token_env_key=None,
                 certificate_request_parameter=None):
        """
        @type resource_server_url: basestring
        @param resource_server_url: URL of resource server from which to request
        certificates
        @type client_cert: basestring
        @param client_cert: location of certificate file for OAuth client
        @type client_key: basestring
        @param client_key: location of private key file for OAuth client
        @type ca_cert_file: basestring
        @param ca_cert_file: location of file containing trusted certificate
        authority certificates
        @type ca_dir: basestring
        @param ca_dir: location of directory containing trusted certificate
        authority certificate files
        @type token_env_key: basestring
        @param token_env_key: key of entry in environ holding the OAuth token
        @type certificate_request_parameter: basestring
        @param certificate_request_parameter: name of the parameter to be used
        in the resource request to hold the certificate request
        """
        self.resource_server_url = resource_server_url
        self.token_env_key = (self.DEFAULT_TOKEN_ENV_KEYNAME
                              if token_env_key is None
                              else token_env_key)

        # SSL configuration
        self.client_cert = client_cert
        self.client_key = client_key
        self.ca_cert_file = ca_cert_file
        self.ca_dir = ca_dir
        self.client_ssl_config = ssl_context_util.SSlContextConfig(client_key,
            client_cert, self.ca_cert_file, self.ca_dir, True)

        # OAuth client configuration
        self.certificate_request_parameter = certificate_request_parameter

        self.token = None
        self.user_ssl_context = None

    def get(self, environ, url):
        """Makes a HTTP request to the specified URL using the certificate
        obtained from the WSGI environ.
        @type environ: dict
        @param environ: WSGI environ
        @type url: basestring
        @param url: URL of resource to request
        @rtype: basestring
        @return: response from HTTP request
        """
        current_token = environ.get(self.token_env_key)
        if current_token:
            log.debug("Token ID: %s", current_token)
        else:
            log.debug("No token found with environ key: %s", self.token_env_key)
            
        if self.token != current_token or not self.user_ssl_context:
            log.debug("Certificate request needed")
            if current_token:
                self.token = current_token
                # Get credential.
                log.debug("Making certificate request")
                (private_key,
                 certificate) = certificate_request.request_certificate(
                                            self.token, 
                                            self.resource_server_url,
                                            self.client_ssl_config, 
                                            self.certificate_request_parameter)
    
                # Create SSL context using the resource owner's delegated
                # credential.
                self.user_ssl_context = ssl_context_util.make_ssl_context(
                                                            None,
                                                            None,
                                                            self.ca_cert_file,
                                                            self.ca_dir,
                                                            True)

                clientKey = crypto.load_privatekey(crypto.FILETYPE_PEM,
                                                   private_key)
                clientCert = crypto.load_certificate(crypto.FILETYPE_PEM, 
                                                     certificate)
   
                self.user_ssl_context.use_privatekey(clientKey)
                self.user_ssl_context.use_certificate(clientCert)
                log.debug("Created new SSL context")
            else:
                log.warn("Certificate needed but no token available")

        config = httpsclientutils.Configuration(self.user_ssl_context, True)
        log.debug("Making request to URL: %s", url)
        response = httpsclientutils.fetch_from_url(url, config)

        return response



"""NDG OAuth 2.0 

Example paster script to run bearer token example client app
"""
__author__ = "Philip Kershaw"
__date__ = "24/08/12"
__copyright__ = "(C) 2012 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id$"
from paste.script.serve import ServeCommand

ServeCommand("serve").run(["./bearer_tok_client_app.ini"])



"""Function to make a request for an X.509 certificate from an Online 
Certificate authority protected with OAuth
"""
__author__ = "R B Wilkinson"
__date__ = "20/03/12"
__copyright__ = "(C) 2012 Science and Technology Facilities Council"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id$"

import base64
import json
import logging
import urllib

from ndg.oauth.client.lib import openssl_cert
from ndg.oauth.client.lib.oauth2client import Oauth2Client

log = logging.getLogger(__name__)

DEFAULT_CERTIFICATE_REQUEST_PARAMETER = 'certificate_request'


def request_certificate(token, 
                        resource_server_url, 
                        ssl_config,
                        certificate_request_parameter=None):
    """Requests a certificate using an OAuth authorized resource request.
    @param token: access token to use in request
    @type token: basestring
    @param resource_server_url: URL of resource server to which the request
    should be made
    @type resource_server_url: basestring
    @param ssl_config: SSL configuration including the OAuth client certificate
    and private key
    @type ssl_config: ndg.httpsclient.ssl_context_util:SSlContextConfig
    @param certificate_request_parameter: name of parameter for the certificate
    request in the resource server request
    @type certificate_request_parameter: basestring
    """
    oauth_client = Oauth2Client(access_token=token)
    
    parameters = {}
    key_pair = openssl_cert.create_keypair()
    cert_req = openssl_cert.create_certreq('ignored-username', key_pair)
    
    cert_req_param = (DEFAULT_CERTIFICATE_REQUEST_PARAMETER
                      if certificate_request_parameter is None
                      else certificate_request_parameter)
    
    parameters[cert_req_param] = base64.b64encode(cert_req)

    # Make POST request to obtain an access token.
    log.debug("Resource request - parameters: %s", parameters)
    data = urllib.urlencode(parameters)
    
    response = oauth_client.request_resource(resource_server_url, 
                                             ssl_config=ssl_config, 
                                             data=data)

    # TODO: Refactor so that does or doesn't support JSON response - currently
    # works so that it will accept either
    if 'application/json' in response.headers.get('Content-type', ''):
        response_json = json.load(response)
        certificate = response_json.get('certificate', None)
    else:       
        certificate = response.read() 

    # Get the private key.
    private_key = openssl_cert.getKeyPairPrivateKey(key_pair)
    return private_key, certificate
